# -*- coding: utf-8 -*-
"""convert to/from orgmode

:copyright: Copyright (c) 2022 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from pykern.pkcollections import PKDict
from pykern.pkdebug import pkdc, pkdlog, pkdp
import pykern.pkcli
import pykern.pkcli.github
import pykern.pkcollections
import pykern.pkio
import re


_PROPERTIES = (
    # order matters, and underscore is a not a GitHub API name
    "assignees",
    "_repo",
    "created_at",
    "html_url",
    "milestone",
    "number",
    "user",
)


def assignee_issues(user, org_d="~/org"):
    """Export issues for auth user to orgmode file named ``org_d/user.org``

    Args:
        user (str): user name to use
        org_d (str): where to store org files [~/org]
    Returns:
        str: Name of org file created
    """
    return str(_OrgModeGen(user=user, org_d=org_d).from_issues())


def from_issues(*repos, org_d="~/org"):
    """Export issues to orgmode file named ``org_d/repo.org``

    Args:
        repos (str): will add radiasoft/ if missing
        org_d (str): where to store org files [~/org]
    Returns:
        str: Name of org file created
    """
    return str(_OrgModeGen(repos=repos, org_d=org_d).from_issues())


def test_data(repo, path):
    """Used to generate the unit test data"""
    from pykern import pkjson

    if len(repo):
        r = pykern.pkcli.github.GitHub().repo_arg(repo)
        res = _dict(r)
    else:
        r = pykern.pkcli.github.GitHub().login()
        res = PKDict()
    res._issues = sorted(
        [_dict(i) for i in r.issues(state="open")],
        key=lambda x: x.number,
    )
    pkjson.dump_pretty(res, filename=path)


def to_issues(org_path, dry_run=False):
    """Import (existing) issues from ``org_d/repo.org``

    Args:
        org_path (str): org mode file
        dry_run (bool): whether to update issues or not
    Returns:
        str: updates made
    """
    return _OrgModeParser(org_path=org_path).to_issues(dry_run)


class _Base:
    _NON_PROPERTIES = ("body", "labels", "title")
    _ATTRS = tuple(k for k in _PROPERTIES + _NON_PROPERTIES if not k.startswith("_"))

    _COMMENT_TAG = ":_separator_:"

    def __init__(self, **kwargs):
        self._github = pykern.pkcli.github.GitHub()

    def _issue_as_dict(self, issue):
        def _str(item):
            if isinstance(item, list):
                return " ".join(sorted(_str(i) for i in item))
            if isinstance(item, str):
                return item
            if item is None:
                return ""
            if isinstance(item, int):
                return str(item)
            if isinstance(item, dict):
                for k in "name", "login", "title":
                    if k in item:
                        return item[k]
            raise AssertionError(f"unknown type={type(item)} item={item}")

        i = _dict(issue)
        return PKDict({k: _str(i[k]) for k in self._ATTRS if k in i})

    def _iter_issues(self, issues):
        for i in issues:
            if not i.pull_request():
                yield i

    def _open_issues(self, repo):
        """Ignores pull requests"""
        return self._iter_issues(repo.issues(state="open"))


class _OrgModeGen(_Base):
    _TITLE = re.compile(r"^(\d{4})-?(\d\d)-?(\d\d)\s*(.*)")
    # strict for now
    _HTML_URL = re.compile(r"https://github.com/([\.\w-]+/[\.\w-]+)/issues/\d+$")
    _NO_DEADLINES_MARK = "ISSUES DO NOT HAVE DEADLINES AFTER THIS " + _Base._COMMENT_TAG
    _CFG = "#+STARTUP: showeverything\n#+COLUMNS: %13DEADLINE %50ITEM %number(Num) %15assignees %TAGS\n"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._org_d = pykern.pkio.py_path(kwargs["org_d"])
        if "repos" in kwargs:
            if not kwargs["repos"]:
                self._error("no repos supplied")
            if len(kwargs["repos"]) == 1:
                r = self._github.repo_arg(kwargs["repos"][0])
                b = f"{r.organization['login']}-{r.name}"
                self._issues = self._open_issues(r)
            else:
                b = "issues"
                self._issues = []
                for r in kwargs["repos"]:
                    self._issues.extend(self._open_issues(self._github.repo_arg(r)))
        elif "user" in kwargs:
            b = kwargs["user"]
            self._issues = self._iter_issues(
                (
                    i.issue
                    for i in self._github.login().search_issues(
                        query=f"assignee:{b} state:open"
                    )
                ),
            )
        else:
            raise AssertionError(f"kwargs={kwargs} invalid")
        self._org_path = self._org_d.join(f"{b}.org")

    def from_issues(self):
        self._no_deadlines = None
        return self._write(self._CFG + "".join(self._issue(i) for i in self._sorted()))

    def _error(self, msg):
        pykern.pkcli.command_error("{} org_d={}", msg, self._org_d)

    def _issue(self, issue):
        def _deadline():
            d = issue.get("_deadline")
            if not d:
                return ""
            return f"DEADLINE: <{d}>\n"

        def _drawer(name, body):
            return f":{name}:\n{body}:END:\n"

        def _properties():
            return _drawer(
                "PROPERTIES",
                "".join(_property(k) for k in _PROPERTIES),
            )

        def _property(name):
            res = f":{name}:"
            if len(issue[name]) > 0:
                res += f" {issue[name]}"
            return res + "\n"

        def _tags():
            if not issue.labels:
                return ""
            return " :" + issue.labels.replace(" ", ":") + ":"

        def _title():
            res = ""
            if self._no_deadlines is None and issue.get("_deadline") is None:
                self._no_deadlines = True
                res = f"* {self._NO_DEADLINES_MARK}\n"
            return f"{res}* {issue.title or ''}{_tags()}\n"

        return _title() + pykern.pkcli.github.GitHub.indent2(
            _deadline()
            + _properties()
            + _drawer("BODY", pykern.pkcli.github.GitHub.issue_body(issue)),
        )

    def _sorted(self):
        repo_name_cache = PKDict()

        def _dict(issue):
            res = self._issue_as_dict(issue)
            m = self._TITLE.search(res.title)
            if m:
                k = res._deadline = f"{m.group(1)}-{m.group(2)}-{m.group(3)}"
                res.title = m.group(4)
            else:
                k = "9999-01-01"
            # created_at makes deterministic
            res._key = f"{k} {res.number:>010} {res.created_at}"
            res._repo = _repo_name(res.html_url)
            return res

        def _repo_name(url):
            return repo_name_cache.pksetdefault(url, lambda: _repo_name_calc(url))[url]

        def _repo_name_calc(url):
            m = self._HTML_URL.search(url)
            if not m:
                raise ValueError(f"html_url={url} invalid format")
            return m.group(1)

        return sorted([_dict(i) for i in self._issues], key=lambda x: x._key)

    def _write(self, text):
        return pykern.pkio.write_text(self._org_path, text)


class _OrgModeParser(_Base):
    _ARRAY_ATTRS = ("assignees", "labels")
    _DEADLINE = re.compile(r"^\s*DEADLINE:\s*<(\d{4})-(\d\d)-(\d\d)")
    _HEADING = re.compile(r"^\*+\s*(.*)")
    #: orgmode indent is always 2; POSIT indent2() indents by 2
    _INDENT = 2
    _PROPERTY = re.compile(r"^:(\w+):\s*(.*)")
    _TAGS = re.compile(r"^(.+)\s+(:(?:\S+:)+)\s*$")

    def __init__(self, org_path):
        super().__init__()
        self._org_path = pykern.pkio.py_path(org_path)

    def to_issues(self, dry_run):
        self._dry_run = dry_run
        self._lines = pykern.pkio.read_text(self._org_path).splitlines()
        self._parse()
        return self._update()

    def _add_issue(self, issue):
        i = self._repos.setdefault(issue._repo, PKDict())
        if issue.number in i:
            self._error(
                f"number={issue.number} duplicated in prev={i[issue.number]} curr={issue}",
            )
        i[issue.number] = issue

    def _body(self, issue):
        issue.body = "\n".join(self._drawer("BODY"))
        if len(issue.body):
            issue.body += "\n"

    def _deadline(self, issue):
        l = self._next("DEADLINE:")
        m = self._DEADLINE.search(l)
        if not m:
            # Optional
            self._lines.insert(0, l)
            return
        issue.title = f"{m.group(1)}{m.group(2)}{m.group(3)} {issue.title}"

    def _drawer(self, name):
        def _line(key):
            x = f":{key}:"
            l = self._next(x)[self._INDENT :]
            if x == l:
                return None
            if key != "END":
                self._error(f"expect={x} but got line={l}")
            return l

        _line(name)
        res = []
        while True:
            l = _line("END")
            if l is None:
                break
            res.append(l)
        return res

    def _error(self, msg):
        pykern.pkcli.command_error("{} path={}", msg, self._org_path)

    def _next(self, expect=None):
        while self._lines:
            res = self._lines.pop(0)
            if res.startswith("#"):
                continue
            return res
        if expect is None:
            return None
        self._error(f"expect={expect} but got EOF")

    def _parse(self):
        self._repos = PKDict()
        while True:
            l = self._next(None)
            if l is None:
                return
            m = self._HEADING.search(l)
            if not m:
                continue
            if self._COMMENT_TAG in m.group(1):
                continue
            self._add_issue(self._parse_issue(l, m.group(1)))

    def _parse_issue(self, line, title):
        res = PKDict()
        self._title(res, title)
        self._deadline(res)
        self._properties(res)
        self._body(res)
        return res

    def _properties(self, issue):
        for l in self._drawer("PROPERTIES"):
            m = self._PROPERTY.search(l)
            if not m:
                self._error(f"expected :property: value but got line={l}")
            issue[m.group(1)] = m.group(2).strip()

    def _title(self, issue, line):
        m = self._TAGS.search(line)
        if m:
            issue.title = m.group(1)
            issue.labels = " ".join(
                sorted(t for t in m.group(2).split(":") if len(t) > 0),
            )
        else:
            issue.title = line
            issue.labels = ""

    def _update(self):
        res = PKDict()
        for k in sorted(self._repos):
            e = self._update_repo(self._github.repo_arg(k), self._repos[k])
            if e:
                res[k] = e
        return res

    def _update_repo(self, repo, issues):
        def _fix_milestone(edits):
            m = edits.get("milestone")
            if m is None:
                return edits
            res = edits.copy()
            try:
                v = self._github.milestone(repo, m)
            except KeyError:
                v = repo.create_milestone(m).number
            res.milestone = v
            return res

        def _edits(base, update):
            res = PKDict()
            for k in "assignees", "body", "labels", "milestone", "title":
                if base[k] != update[k]:
                    res[k] = (
                        sorted(update[k].split())
                        if k in self._ARRAY_ATTRS
                        else update[k]
                    )
                    if k == "labels":
                        y = set(res[k]) - set(x.name for x in repo.labels())
                        if y:
                            # labels will be created automatically, which is unlikely
                            # something we want, since labels have colors and we
                            # have a tool to create labels.
                            raise ValueError(f"non-existent labels={y}")
            return res

        res = PKDict()
        # only update issues that are still open
        for i in self._open_issues(repo):
            try:
                u = issues.get(str(i.number))
                if not u:
                    continue
                e = None
                e = _edits(self._issue_as_dict(i), u)
                if e and not self._dry_run:
                    i.edit(**_fix_milestone(e))
                if e:
                    res[i.number] = e
            except Exception:
                pkdlog("edits={} error in issue={}", e, f"{repo}#{i.number}")
                raise
        return res


def _dict(model):
    return pykern.pkcollections.canonicalize(model.as_dict())



from pykern.pkcollections import PKDict
x=PKDict(foo='foo', bar='bar', a='a', b='b', c='c', d='d', e='e', f='f', param1='foo', param2='bar', foobar='foobar', barfoo='barfoo')
y='''
asdfasdfasdfasdf
'''
class foo:
    def __init__(self, a, b, c):
        self.a = a
        self.b = b
        self.c=f'{c}'
    def get_c(self):
        return self.c




g=foo(x, y, 3)



# -*- coding: utf-8 -*-
u"""pksetupunit2 setup script

:copyright: Copyright (c) 2016 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
import pykern.pksetup

pykern.pksetup.setup(
    name='pksetupunit2',
    author='RadiaSoft LLC',
    author_email='pip@radiasoft.net',
    description='optional args like extras_require',
    license='http://www.apache.org/licenses/LICENSE-2.0.html',
    url='https://github.com/radiasoft/pksetupunit2',
    extras_require={
        # Something that won't have already been installed
        'r1': ['adhan'],
        'r2': ['civicjson'],
    },
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python',
        'Topic :: Utilities',
    ],
)



"""pytest for `pykern.pkcli`

:copyright: Copyright (c) 2015-2023 Bivio Software, Inc.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

import pytest


def test_argh_argument_parsing(capsys):
    """Test change in how kwargs are handled.

    Prior to argh v0.30.0 keyword_arg would be an optional flag
    argument So, you could have `arghparsing argh-test kwarg-to-positional x`
    or `pykern argh-test kwarg-to-positional x --keyword-arg foo`.

    In version >= 0.30.0 keyword_arg became an optional positional
    arg. So, you could have `pykern argh-test kwarg-to-positional x`
    or `pykern argh-test kwarg-to-positional x foo`
    But, if you set name_mapping_policy to BY_NAME_IF_HAS_DEFAULT then
    the old bheavior is retained.

    See the argh changelog for more information:
    https://argh.readthedocs.io/en/latest/changes.html#version-0-30-0-2023-10-21
    """
    from pykern.pkunit import pkre, pkeq

    p = "p"
    k = "k"
    pkeq(
        _main(
            "package2",
            [
                "argh_test",
                "kwarg-to-positional",
                p,
                "--keyword-arg",
                k,
            ],
        ),
        0,
    )
    pkre(f"positional_arg={p} keyword_arg={k}", capsys.readouterr()[0])


def test_command_error(capsys):
    from pykern import pkcli
    from pykern import pkconfig

    pkconfig.reset_state_for_testing()
    with pytest.raises(pkcli.CommandError) as e:
        pkcli.command_error("{abc}", abc="abcdef")
    assert "abcdef" in str(
        e.value
    ), "When passed a format, command_error should output formatted result"
    _deviance(
        "package2", ["some-mod", "command-error"], None, r"raising CommandError", capsys
    )


def test_main1():
    """Verify basic modes work"""
    from pykern import pkconfig

    pkconfig.reset_state_for_testing()
    rp = "package1"
    _conf(rp, ["conf1", "cmd1", "1"])
    _conf(rp, ["conf1", "cmd2"], first_time=False)
    _conf(rp, ["conf2", "cmd1", "2"])
    _conf(rp, ["conf3", "3"], default_command=True)


def test_main2(capsys):
    from pykern import pkconfig
    import six

    all_modules = r":\nconf1\nconf2\nconf3\n$"
    pkconfig.reset_state_for_testing()
    rp = "package1"
    _deviance(rp, [], None, all_modules, capsys)
    _deviance(rp, ["--help"], None, all_modules, capsys)
    _deviance(rp, ["conf1"], SystemExit, r"cmd1,cmd2.*too few", capsys)
    _deviance(rp, ["conf1", "-h"], SystemExit, r"\{cmd1,cmd2\}.*commands", capsys)
    if six.PY2:
        _deviance(rp, ["not_found"], None, r"no module", capsys)
    else:
        _deviance(rp, ["not_found"], ModuleNotFoundError, None, capsys)
    _deviance(rp, ["conf2", "not-cmd1"], SystemExit, r"\{cmd1\}", capsys)


def test_main3():
    """Verify underscores are converted to dashes"""
    from pykern import pkconfig

    pkconfig.reset_state_for_testing()
    assert 0 == _main(
        "package2", ["some-mod", "some-func"]
    ), "some-mod some-func: dashed module and function should work"
    assert 0 == _main(
        "package2", ["some_mod", "some_func"]
    ), "some_mod some-func: underscored module and function should work"


def test_command_info():
    from pykern import pksubprocess
    from pykern.pkunit import case_dirs

    for d in case_dirs("command_info"):
        pksubprocess.check_call_with_signals(
            ["python", d.join("example_pkcli_module.py")],
            output="example_stderr.out",
        )


def _conf(root_pkg, argv, first_time=True, default_command=False):
    import sys

    full_name = ".".join([root_pkg, "pkcli", argv[0]])
    if not first_time:
        assert not hasattr(sys.modules, full_name)
    assert _main(root_pkg, argv) == 0, "Unexpected exit"
    m = sys.modules[full_name]
    if default_command:
        assert m.last_cmd.__name__ == "default_command"
        assert m.last_arg == argv[1]
    else:
        assert m.last_cmd.__name__ == argv[1]


def _deviance(root_pkg, argv, exc, expect, capsys):
    import re
    from pykern.pkdebug import pkdp
    from pykern import pkunit

    if exc:
        with pytest.raises(exc):
            _main(root_pkg, argv)
        if not expect:
            return
    else:
        assert _main(root_pkg, argv) == 1, f"Failed to exit(1): {argv}"
    out, err = capsys.readouterr()
    if not err:
        err = out
    assert (
        re.search("Args.*arg1", err, flags=re.DOTALL) is None
    ), "failure to ignore arguments and only print subject. out: {}".format(err)
    pkunit.pkre(expect, err)


def _main(root_pkg, argv):
    import sys
    from pykern import pkunit, pkcli

    sys.argv[:] = ["pkcli_test"]
    sys.argv.extend(argv)
    dd = str(pkunit.data_dir())
    try:
        sys.path.insert(0, dd)
        return pkcli.main(root_pkg)
    finally:
        if sys.path[0] == dd:
            sys.path.pop(0)



"""PyTest for :mod:`pykern.pkyaml`

:copyright: Copyright (c) 2015-2022 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""


def test_load_dump():
    """Test values are unicode"""
    from pykern import pkunit
    from pykern import pkyaml

    for d in pkunit.case_dirs():
        y = pkyaml.load_file(d.join("in.yml"))
        _assert_load(y)
        pkyaml.dump_pretty(y, d.join("out.yml"))


def test_load_resource():
    """Test file can be read"""
    from pykern import pkunit
    from pykern import pkyaml

    p1 = pkunit.import_module_from_data_dir("p1")
    pkunit.pkeq(
        "v2",
        p1.y["f2"],
        "Resource should be loaded relative to root package of caller",
    )


def _assert_load(value):
    from pykern.pkunit import pkok, pkfail
    from pykern.pkcollections import PKDict

    if isinstance(value, PKDict):
        for k, v in value.items():
            _assert_load(k)
            _assert_load(v)
    elif isinstance(value, list):
        for v in value:
            _assert_load(v)
    else:
        pkok(
            isinstance(value, (int, float, str)),
            "unknown type={} value={}",
            type(value),
            value,
        )



"""PyTest for :mod:`pykern.pkio`

:copyright: Copyright (c) 2015 Bivio Software, Inc.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

import glob
import os
import py
import pytest


def test_atomic_write():
    from pykern import pkunit
    from pykern import pkio

    with pkunit.save_chdir_work():
        pkio.atomic_write("x.ABC", "abc")
        pkunit.pkeq("abc", pkio.read_text("x.ABC"))


def test_compare_files():
    from pykern import pkio
    from pykern import pkunit
    from pykern.pkunit import pkok

    with pkunit.save_chdir_work():
        text = "abc"
        b = pkio.py_path("base")
        b.write(text)
        c = pkio.py_path("copy")
        b.copy(c)
        c.setmtime(b.mtime())
        c = pkio.py_path("same-content")
        c.write(text)
        c.setmtime(b.mtime() + 10)
        c = pkio.py_path("same-stat")
        c.write(len(text) * "X")
        c.setmtime(b.mtime())
        c = pkio.py_path("diff-size")
        c.write(text + text)
        c.setmtime(b.mtime())
        pkok(pkio.compare_files("base", "copy"), "copy")
        pkok(pkio.compare_files("base", "same-content"), "same-content")
        pkok(pkio.compare_files("base", "same-stat"), "same-stat")
        pkok(not pkio.compare_files("base", "same-stat", force=True), "same-stat force")
        pkok(not pkio.compare_files("base", "diff-size"), "diff-size")
        pkok(not pkio.compare_files("base", "not-found"), "not-found")
        pkok(not pkio.compare_files("both-not-found", "not-found"), "both-not-found")


def test_has_file_extension():
    from pykern.pkunit import pkeq
    from pykern import pkio

    pkeq(True, pkio.has_file_extension("x.ABC", "abc"))
    pkeq(True, pkio.has_file_extension(py.path.local("x.abc"), ("abc", "def")))
    pkeq(False, pkio.has_file_extension("filename_with_no_extension", "json"))


def test_is_pure_text():
    from pykern import pkunit
    from pykern import pkio

    d = pkunit.data_dir()
    pkunit.pkeq(False, pkio.is_pure_text(d.join("binary.dat")))
    pkunit.pkeq(True, pkio.is_pure_text(d.join("text.dat")))


def test_py_path():
    from pykern import pkunit
    from pykern import pkio
    from pykern.pkunit import pkeq

    with pkunit.save_chdir_work():
        d = pkunit.data_dir()
        pkeq(d, pkio.py_path(d))


def test_save_chdir():
    from pykern import pkunit
    from pykern import pkio

    expect_prev = py.path.local().realpath()
    expect_new = py.path.local("..").realpath()
    try:
        with pkio.save_chdir(expect_new) as new:
            assert (
                expect_new == new
            ), "save_chdir returns current directory before chdir"
            assert (
                expect_new == py.path.local().realpath()
            ), "When in save_chdir, expect current directory to be new directory"
            os.chdir("..")
            assert (
                expect_new != py.path.local().realpath()
            ), "When in save_chdir, expect chdir to still work"
            raise IndentationError()
    except IndentationError as benign_exception:
        pass
    assert (
        expect_prev == py.path.local().realpath()
    ), "When exception is raised, current directory should be reverted."
    expect_new = pkunit.empty_work_dir().join("new_folder").realpath()
    with pytest.raises(OSError):
        with pkio.save_chdir(expect_new) as new:
            assert (
                False
            ), "When save_chdir given non-existent dir, should throw exception"
    with pkio.save_chdir(expect_new, mkdir=True) as new:
        assert (
            expect_new == py.path.local().realpath()
        ), "When save_chdir given non-existent dir and mkdir=True, should pass"


def test_unchecked_remove():
    """Also tests mkdir_parent"""
    from pykern import pkunit
    from pykern import pkio

    with pkunit.save_chdir_work():
        fn = "f1"
        # Should not throw an exception
        pkio.unchecked_remove(fn)
        pkio.write_text(fn, "hello")
        pkio.unchecked_remove(fn)
        assert not os.path.exists(fn), "When file removed, should be gone"
        for f in ("d1", "d2/d3"):
            assert py.path.local(f) == pkio.mkdir_parent(
                f
            ), "When mkdir_parent is called, returns path passed in"
        assert os.path.exists("d1"), "When single directory, should exist"
        assert os.path.exists("d2/d3"), "When nested directory, should exist"
        with pytest.raises(AssertionError):
            pkio.unchecked_remove(".")
        with pytest.raises(AssertionError):
            pkio.unchecked_remove(os.getcwd())
        with pytest.raises(AssertionError):
            pkio.unchecked_remove("/")


def test_walk_tree_and_sorted_glob():
    """Looks in work_dir"""
    from pykern import pkunit
    from pykern import pkio
    import re

    with pkunit.save_chdir_work() as pwd:
        for f in ("d1/d7", "d2/d3", "d4/d5/d6"):
            pkio.mkdir_parent(f)
        expect = []
        for f in ["d1/d7/f1", "d4/d5/f2", "d2/d3/f3"]:
            pkio.write_text(f, "")
            expect.append(py.path.local(f))
        assert sorted(expect) == list(
            pkio.walk_tree(".")
        ), "When walking tree, should only return files"
        assert [expect[2]] == list(
            pkio.walk_tree(".", "f3")
        ), "When walking tree with file_re, should only return matching files"
        assert [expect[0]] == list(
            pkio.walk_tree(".", "^d1")
        ), "When walking tree with file_re, file to match does not include dir being searched"
        assert [expect[0]] == list(
            pkio.walk_tree(".", re.compile("^d1"))
        ), "When walking tree with file_re, file to match does not include dir being searched"
        assert pkio.sorted_glob("*/*/f*", key="basename") == expect


def test_write_binary():
    """Also tests read_binary"""
    from pykern import pkunit
    from pykern import pkio

    d = pkunit.empty_work_dir()
    expect_res = d.join("anything")
    expect_content = b"\xef broken utf8 \xbb\xbf"
    res = pkio.write_binary(expect_res, expect_content)
    pkunit.pkeq(expect_res, res)
    pkunit.pkeq(expect_content, pkio.read_binary(expect_res))


def test_write_text():
    """Also tests read_text"""
    from pykern import pkunit
    from pykern import pkio

    d = pkunit.empty_work_dir()
    expect_res = d.join("anything")
    expect_content = "something\u2167"
    write_content = bytes(expect_content, "utf-8")
    res = pkio.write_text(str(expect_res), write_content)
    assert expect_res == res, "Verify result is file path as py.path.Local"
    with open(str(expect_res), "rb") as f:
        assert (
            write_content == f.read()
        ), 'When write_text is called, it should write "something"'
    assert expect_content == pkio.read_text(
        str(expect_res)
    ), 'When read_text, it should read "something"'
    pkio.write_binary(expect_res, b"\xFF\xFF\0")
    with pkunit.pkexcept(str(expect_res)):
        pkio.read_text(expect_res)
    with pkunit.pkexcept(str(expect_res)):
        pkio.write_text(expect_res, b"\xFF\xFF\0")



"""test pytest skip

:copyright: Copyright (c) 2024 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
import pytest


@pytest.mark.skip(reason="always skips")
def test_skip():
    from pykern import pkunit

    pkunit.pkfail("test_skip not skipped")


def test_pass():
    pass



# -*- coding: utf-8 -*-
"""pkconfig init

:copyright: Copyright (c) 2019 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
import pytest


def test_init(pkconfig_setup):
    """Validate parse_set"""
    pkconfig = pkconfig_setup(
        cfg=dict(
            PYKERN_PKCONFIG_CHANNEL="alpha",
            P1_M1_SET3="",
            P1_M1_SET4="a:b",
        ),
        env=dict(P1_M1_REQ8="99", P1_M1_REQ10="h1:h2"),
    )
    from pykern import pkunit
    from p1.m1 import cfg

    pkunit.pkeq(("h1", "h2"), cfg.req10)
    pkunit.pkeq(False, pkconfig.in_dev_mode())



"""Useful operations for unit tests

:copyright: Copyright (c) 2015 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

from pykern import pkcompat
from pykern import pkinspect
from pykern import pkio

# defer importing pkconfig
import pykern.pkconst
import contextlib
import importlib
import inspect
import json
import os
import py
import pytest
import random
import re
import socket
import subprocess
import sys
import traceback

#: Environment var set by pykern.pkcli.test for each module under test
TEST_FILE_ENV = "PYKERN_PKUNIT_TEST_FILE"

#: Environment var set by pykern.pkcli.test if the test is restartable
RESTARTABLE = "PYKERN_PKUNIT_RESTARTABLE"

#: Where persistent input files are stored (test_base_name_data)
DATA_DIR_SUFFIX = "_data"

#: Used to create test servers
LOCALHOST_IP = "127.0.0.1"

#: Where to write temporary files (test_base_name_work)
WORK_DIR_SUFFIX = "_work"

#: Where `ExceptToFile` writes exception
PKEXCEPT_PATH = "pkexcept"

#: Where `ExceptToFile` writes stack
PKSTACK_PATH = "pkstack"

#: INTERNAL: Set to the most recent test module by `_test_file`
module_under_test = None

#: Type of a regular expression
_RE_TYPE = type(re.compile(""))

#: used by case_dirs for comparing sheets
_CSV_SHEET_ID = re.compile(r"(.+)#(\d)$")

#: _test_file initialized?
_init_test_file = False

#: module being run by `pykern.pkcli.test`
__test_file = None


class PKFail(AssertionError):
    pass


class ExceptToFile:
    """Writes exception or None to `PKEXCEPT_PATH`

    Used for deviance testing with `case_dirs`.

    If there is an exception, writes that to the file. Otherwise, writes "None"

    If there is an exception, will write `PKSTACK_PATH`. Otherwise, no
    file exists. Used for diagnostics only.

    Usage::
        for d in case_dirs():
            with ExceptToFile():
                command to test

    Returns:
        None: just for context manager
    """

    def __enter__(self):
        return None

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            r = str(None)
        else:
            r = re.sub(
                pkio.py_path().dirname + r"\S*/", "", str(exc_val), flags=re.IGNORECASE
            )
            with open(PKSTACK_PATH, "wt") as f:
                traceback.print_exception(exc_type, exc_val, exc_tb, file=f)
        pkio.write_text(PKEXCEPT_PATH, r + "\n")
        return True


def assert_object_with_json(
    basename,
    actual,
):
    """Converts actual to JSON and compares with data_dir/basename.json

    Reads data_dir/basename.json and compares with actual
    converted to json. Trailing newline is managed properly. The
    keys are sorted and indentation is 4. actual written to work_dir.

    Args:
        expected_basename (str): file to be found in data_dir with json suffix
        actual (object): to be serialized as json
    """
    from pykern.pkdebug import pkdpretty

    actual = pkdpretty(actual)
    fn = "{}.json".format(basename)
    a = work_dir().join(fn)
    pkio.write_text(a, actual)
    e = data_dir().join(fn)
    expect = pkio.read_text(e)
    pkeq(expect, actual, "diff {} {}", e, a)


def case_dirs(group_prefix="", **kwargs):
    """Sets up `work_dir` by iterating ``*.in`` in `data_dir`

    Every ``<case-name>.in`` is copied recursively to ``<case-name>`` in
    `work_dir`. This function then yields that directory. The test can
    then run the function to be tested.

    When test yields to the generator, this function looks for all
    files in `data_dir` in the sub-directory ``<case-name>.out``. Each
    of these expect files is copmared to the corresponding `work_dir`
    actual file using `file_eq`.

    If you want to only use cases from some specific `<case-name>.in`
    subdir, and not all `*.in` subdirs, you can pass a `group_prefix`
    default parameter value ('' by default) to `case_dirs()`. This will
    perform the regular operations but only on `<case-name>.in`.

    Excel spreadsheets are supported. If you want to automatically
    compare xlsx files, you need to install ``pandas``, which will be
    used to convert Excel files as follows. If the name of the expect
    (out) file is ``foo.csv``, then the first sheet (sheet 0) in the
    corresponding work_dir xlsx will be converted to ``foo.csv``
    before comparison.  If the expect (out) file has a ``#<digit>``,
    e.g. ``foo#3.csv``, then the fourth sheet will be extracted from
    the actual xlsx to ``foo#3.csv`` in the work_dir.

    If ExceptToFile in the body of a case_dirs loop, the exception
    will be output if a file is not found.

    Args:
        group_prefix (string): target subdir ['']
        j2_ctx (dict): passed to `pykern.pkjinja.render_file`
        ignore_lines (iterable): `POSIX standard regular expressions <https://www.gnu.org/software/findutils/manual/html_node/find_html/posix_002dbasic-regular-expression-syntax.html>`_ to be passed to `diff`
        is_bytes (bool): do a binary comparison [False]

    Returns:
        py.path.local: case directory created in work_dir (also PWD)

    """
    import shutil

    def _compare(in_d, work_d):
        o = in_d.new(ext="out")
        for e in pkio.walk_tree(o):
            if e.basename.endswith("~"):
                continue
            a = work_d.join(o.bestrelpath(e))
            file_eq(
                expect_path=e,
                actual_path=a,
                **kwargs,
            )

    d = work_dir()
    n = 0
    for i in pkio.sorted_glob(data_dir().join(group_prefix + "*.in")):
        w = d.join(i.purebasename)
        shutil.copytree(str(i), str(w))
        n += 1
        with pkio.save_chdir(w):
            _pkdlog("case_dir={}", i.basename)
            yield w
        try:
            _compare(i, w)
            continue
        except Exception as e:
            # Not found indicates expected output not found.
            # It might have been caused by an exception which was
            # caught by ExceptToFile.
            if not pkio.exception_is_not_found(e):
                raise
            f = w.join(PKSTACK_PATH)
            if not f.exists():
                raise
            _pkdlog("Exception in case_dir={}\n{}", w, f.read())
        # This avoids confusing "during handling of above exception"
        pkfail("See stack above")
    if n == 0:
        pkfail(f"No files found for group_prefix={group_prefix}")


def data_dir():
    """Compute the data directory based on the test name

    The test data directory is always ``<test>_data``, where ``<test>``
    is the name of the test's python module with the ``_test`` or
    ``test_`` removed.  For example, if the test file is
    ``setup_test.py`` then the directory will be ``setup_data``.

    Returns:
        py.path.local: data directory

    """
    return _base_dir(DATA_DIR_SUFFIX)


def data_yaml(base_name):
    """Load base_name.yml from data_dir

    Args:
        base_name (str): name of YAML file with ``.yml`` extension

    Returns:
        object: YAML data structure, usually dict or array
    """
    from pykern import pkyaml

    return pkyaml.load_file(data_dir().join(base_name) + ".yml")


def empty_work_dir():
    """Remove `work_dir` if it exists and create.

    All contents of the test directory will be removed.

    Returns:
        py.path.local: empty work directory

    """
    d = work_dir()
    if os.path.exists(str(d)):
        # doesn't ignore "not found" errors
        d.remove(rec=1, ignore_errors=True)
    return d.ensure(dir=True)


def file_eq(expect_path, *args, **kwargs):
    """If actual is not expect_path, throw assertion with calling context.

    `expect_path` and `actual_path` both exist, they will be compared as plain text.

    If `actual_path` does not exist, it will be created from `actual`.

    If `expect_path` ends in ``.json`` and `actual_path` does not exist,
    `pkjson` will be used to load `expect_path` and a data structure comparison
    will be used with `actual` (and `actual_path` will be written.
    This allows easy testing of complex results.

    If `expect_path` ends with ``.jinja``, it will be rendered
    with `pykern.pkjina.render_file`, and you must supply `j2_ctx`
    in kwargs.

    Args:
        expect_path (str or py.path): text file to be read; if str, then joined with `data_dir`
        actual (object): string or json data structure; if missing, read `actual_path` (may be positional)
        actual_path (py.path or str): where to write results; if str, then joined with `work_dir`; if None, ``work_dir().join(expect_path.relto(data_dir()))``
        j2_ctx (dict): passed to `pykern.pkjinja.render_file`
        ignore_lines (iterable): `POSIX standard regular expressions <https://www.gnu.org/software/findutils/manual/html_node/find_html/posix_002dbasic-regular-expression-syntax.html>`_ to be passed to `diff`
        is_bytes (bool): do a binary comparison [False]
    """
    _FileEq(expect_path, *args, **kwargs)


def unbound_localhost_tcp_port(start, stop):
    """Looks for AF_INET SOCK_STREAM port for which bind succeeds

    Args:
        start (int): first port
        stop (int): one greater than last port (passed to range)
    Returns:
        int: port is available or raises ValueError
    """

    def _check_port(port):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.bind((LOCALHOST_IP, int(port)))
        return port

    for p in random.sample(range(start, stop), 100):
        try:
            return _check_port(p)
        except Exception:
            pass
    raise ValueError(
        f"unable find port random sample range={start}-{stop} tries=100 ip={LOCALHOST_IP}"
    )


def is_test_run():
    """Running in a test?

    Returns:
        bool: whether this is running in a test
    """
    return bool(_test_file())


def import_module_from_data_dir(module_name):
    """Add `data_dir` to sys.path and import module_name.

    Note that `module_name` with be removed from the sys.modules cache
    before loading in case the module was loaded by another test.

    Args:
        module_name (str): module relative to `data_dir` to import.

    Returns:
        module: imported module
    """
    d = str(data_dir())
    prev_path = sys.path
    try:
        sys.path = [d]
        try:
            del sys.modules[module_name]
        except KeyError:
            pass
        m = importlib.import_module(module_name)
        return m
    finally:
        sys.path = prev_path


def pkeq(expect, actual, *args, **kwargs):
    """If actual is not expect, throw assertion with calling context.

    Opposite of `pkne`.

    Args:
        expect (object): what to test for
        actual (object): run-time value
        args (tuple): passed to pkfail()
        kwargs (dict): passed to pkfail()
    """
    if expect != actual:
        _fail(("expect={} != actual={}", expect, actual), *args, **kwargs)


@contextlib.contextmanager
def pkexcept(exc_or_re, *args, **kwargs):
    """Expect an exception to be thrown and match or output msg

    Examples::

        # Expect an exception (or its subclass)
        with pkexcept(AssertionError, 'did not expect this'):
            assert 0

        # Expect exception to contain a specific message
        with pkexcept('match this', 'problem with matching'):
            assert 0, 'some string with "match this" in it'

        # Use a default output message
        with pkexcept(KeyError):
            something['key will not be found']

    Args:
        exc_or_re (object): BaseException, re, or str; if str, compiled with `re.IGNORECASE`
        args (tuple): passed to format
        kwargs (dict): passed to format

    Yields:
        None: just for context manager
    """
    try:
        yield None
    except BaseException as e:
        from pykern.pkdebug import pkdexc

        e_str = "{} {}".format(type(e), e)
        if isinstance(exc_or_re, type) and issubclass(exc_or_re, BaseException):
            if isinstance(e, exc_or_re):
                return
            m = (
                "{}: an exception was raised, but expected it to be {}; stack={}",
                e_str,
                exc_or_re,
                pkdexc(),
            )
        else:
            if not isinstance(exc_or_re, _RE_TYPE):
                exc_or_re = re.compile(exc_or_re, flags=re.IGNORECASE)
            if exc_or_re.search(e_str):
                return
            m = (
                '{}: an exception was raised, but did not match "{}"; stack={}',
                e_str,
                exc_or_re.pattern,
                pkdexc(),
            )
    else:
        m = ("Exception was not raised: expecting={}", exc_or_re)
    _fail(m, *args, **kwargs)


def pkfail(fmt, *args, **kwargs):
    """Format message and raise PKFail.

    Args:
        fmt (str): to be passed to `string.format`
        args (tuple): passed to format
        kwargs (dict): passed to format
    """
    msg = fmt.format(*args, **kwargs)
    call = pkinspect.caller(ignore_modules=[contextlib])
    raise PKFail("{} {}".format(call, msg))


def pkne(expect, actual, *args, **kwargs):
    """If actual is equal to expect, throw assertion with calling context

    Opposite of `pkeq`.

    Args:
        expect (object): what to test for
        actual (object): run-time value
        args (tuple): passed to pkfail()
        kwargs (dict): passed to pkfail()
    """
    if expect == actual:
        _fail(("expect={} == actual={}", expect, actual), *args, **kwargs)


def pkok(cond, fmt, *args, **kwargs):
    """If cond is not true, throw PKFail with calling context

    Args:
        cond (object): expression which should evaluate to true
        fmt (str): to be passed to `string.format`
        args (tuple): passed to format
        kwargs (dict): passed to format

    Returns:
        object: `obj` value
    """
    if not cond:
        pkfail(fmt, *args, **kwargs)
    return cond


def pkre(expect_re, actual, flags=re.IGNORECASE + re.DOTALL):
    """If actual does not match (re.search) expect_re, throw PKFail with calling context.

    Args:
        expect_re (object): string or re object
        actual (object): run-time value
        flags: passed on to re.search [IGNORECASE + DOTALL]
    """
    if not re.search(expect_re, pkcompat.from_bytes(actual), flags=flags):
        pkfail("expect_re={} != actual={}", expect_re, actual)


def restart_or_fail(*args, **kwargs):
    """Test will be restarted (at process level) if it can, else `pkfail`

    Called by tests which experience known CI failures such
    as not being able to connect to servers.

    Communicates with pykern.pkcli.test

    Args:
        fmt (str): to be passed to `string.format`
        args (tuple): passed to format
        kwargs (dict): passed to format
    """
    if os.environ.get(RESTARTABLE):
        raise KeyboardInterrupt()
    pkfail(*args, **kwargs)


def save_chdir_work(is_pkunit_prefix=False, want_empty=True):
    """Change to `work_dir` which will be created.

    Default to `empty_work_dir` before chdir.

    Args:
        is_pkunit_prefix (bool): use as root of (most) file I/O (optional)
        want_empty (bool): call `empty_work_dir` before chdir if True [True]

    Returns:
        py.path.local: empty work directory

    """
    return pkio.save_chdir(
        empty_work_dir() if want_empty else work_dir(),
        is_pkunit_prefix=is_pkunit_prefix,
    )


def work_dir():
    """Returns ephemeral work directory, created if necessary.

    To enable easier debugging, the test directory is always
    ``<test>_work``, where ``<test>`` is the name of the test's python
    module with the ``_test`` or ``test_`` removed.  For example, if the
    test file is ``setup_test.py`` then the directory will be
    ``setup_work``.

    The name "work" distinguishes from "tmp", which could imply
    anything. Also, with editor autocomplete, "setup_work" and
    "setup_test" are more easily distinguishable.

    Returns:
        py.path: directory name
    """
    return _base_dir(WORK_DIR_SUFFIX).ensure(dir=True)


class _FileEq:
    """Implements `file_eq`"""

    def __init__(self, expect_path, *args, **kwargs):
        self._validate_args(expect_path, *args, **kwargs)
        self._set_expect_and_actual()
        self._compare()

    def _actual_xlsx(self):
        try:
            b = self._actual_path.new(ext=".xlsx")
            m = _CSV_SHEET_ID.search(b.purebasename)
            s = 0
            if m:
                b = b.new(purebasename=m.group(1))
                s = int(m.group(2))
            if b.check(file=True):
                self._actual_xlsx_to_csv(b, s)

                return True
            return False
        except Exception:
            _pkdlog(
                "ERROR converting xlsx to csv expect={} actual={}",
                self._expect_path,
                self._actual_path,
            )
            raise

    def _actual_xlsx_to_csv(self, actual_xlsx, sheet):
        import pandas

        p = pandas.read_excel(
            actual_xlsx,
            index_col=None,
            sheet_name=sheet,
        )
        p.columns = p.columns.map(lambda c: "" if "Unnamed" in str(c) else str(c))
        p.to_csv(
            str(self._actual_path),
            encoding="utf-8",
            index=False,
            lineterminator="\r\n",
        )

    def _compare(self):
        w = work_dir()

        def _cmd():
            if self.is_bytes:
                r = ["cmp"]
            else:
                r = ["diff"]
                for l in self._ignore_lines or ():
                    r.extend(("-I", l))
            return r + [str(self._expect_path), str(self._actual_path)]

        def _failed_msg(process):
            r = "'" + "' '".join(process.args) + f"'\n" + process.stdout + "\n"
            if not (process.returncode == 1 or self.is_bytes):
                return r + "diff command failed\n"
            if self._expect_is_jinja:
                return (
                    r
                    + f"""
Implementation restriction: expect is a jinja template which has been processed to
produce the diff. A simple copy of actual to expect is not possible. You will need to update
the expect jinja template={self._expect_path} manually.
"""
                )
            return r + self._update_message

        def _ndiff_config(work_d):
            return pykern.pkio.write_text(
                work_d.join("ndiff_conf.txt"),
                f"* * {'abs' if self._ndiff_epsilon_is_abs else 'rel'}={float(self._ndiff_epsilon)}",
            )

        def _ndiff_files(expect_path, actual_path):
            p = subprocess.run(
                (
                    "ndiff",
                    actual_path,
                    expect_path,
                    _ndiff_config(w),
                ),
                stderr=subprocess.PIPE,
                stdout=subprocess.PIPE,
            )
            d = pkcompat.from_bytes(p.stderr)
            if not re.search(r"processing '.*'\n\s*\d+ lines have been diffed\s*$", d):
                pkfail("diffs detected: {} {}", d, self._update_message)

        if self._is_ndiff:
            _ndiff_files(
                self._expect_path,
                self._actual_path,
            )
            return
        if self._expect == self._actual:
            return
        p = subprocess.run(
            _cmd(), stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True
        )
        if p.returncode != 0:
            pkfail("expect != actual:\n{}", _failed_msg(p))

    def _expect_csv(self):
        if not self._expect_path.ext == ".csv":
            return False
        self._actual_xlsx()
        self._actual = pkio.read_text(self._actual_path)
        self._expect = pkio.read_text(self._expect_path)
        return True

    def _expect_default(self):
        self._expect = self._read(self._expect_path)
        if self._have_actual_kwarg:
            self._write(self._actual_path, self._actual)

    def _expect_jinja(self):
        if not self._expect_is_jinja:
            return False
        import pykern.pkjinja

        self._expect = pykern.pkjinja.render_file(
            self._expect_path, self.j2_ctx, strict_undefined=True
        )
        if self._have_actual_kwarg:
            pkio.write_text(self._actual_path, self._actual)
        return True

    def _expect_json(self):
        if not self._expect_path.ext == ".json" or self._actual_path.exists():
            return False
        self._expect = pkio.read_text(self._expect_path)
        if self._have_actual_kwarg:
            import pykern.pkjson

            pkio.mkdir_parent_only(self._actual_path)
            self._actual = pykern.pkjson.dump_pretty(
                self._actual, filename=self._actual_path
            )
        return True

    def _set_expect_and_actual(self):
        self._read, self._write = (
            (pkio.read_binary, pkio.write_binary)
            if self.is_bytes
            else (pkio.read_text, pkio.write_text)
        )
        self._update_message = f"""
to update test data:
        cp '{self._actual_path}' '{self._expect_path}'
"""
        if self._expect_csv():
            assert not self.is_bytes, "csv is not compatible with is_bytes"
            return
        if self._have_actual_kwarg:
            self._actual = self.kwargs["actual"]
            if self._actual_path.exists():
                pkfail(
                    "actual={} and actual_path={} both exist",
                    self._actual,
                    self._actual_path,
                )
        else:
            self._actual = self._read(self._actual_path)
        if self._expect_json() or self._expect_jinja():
            assert not self.is_bytes, "json or jinja is not compatible with is_bytes"
            return
        self._expect_default()

    def _validate_args(self, expect_path, *args, **kwargs):
        from pykern.pkcollections import PKDict

        self.kwargs = kwargs
        self._have_actual_kwarg = "actual" in self.kwargs
        if args:
            assert (
                not self._have_actual_kwarg
            ), f'have actual as positional arg={args[0]} and kwargs={self.kwargs["actual"]}'
            assert (
                len(args) == 1
            ), f"too many positional args={args}, may only have one (actual)"
            self.kwargs["actual"] = args[0]
            self._have_actual_kwarg = True
        self._actual_path = kwargs.get("actual_path")
        self._expect_path = expect_path
        if not isinstance(self._expect_path, pykern.pkconst.PY_PATH_LOCAL_TYPE):
            self._expect_path = data_dir().join(self._expect_path)
        self._expect_is_jinja = self._expect_path.ext == ".jinja"
        self._is_ndiff = self._expect_path.ext == ".ndiff"
        self._ndiff_epsilon = kwargs.get("ndiff_epsilon", 1e-13)
        self._ndiff_epsilon_is_abs = kwargs.get("ndiff_epsilon_is_abs", False)
        b = (
            self._expect_path.purebasename
            if self._expect_is_jinja
            else self._expect_path.relto(data_dir())
        )
        if self._actual_path is None:
            self._actual_path = b
        if not isinstance(self._actual_path, pykern.pkconst.PY_PATH_LOCAL_TYPE):
            self._actual_path = work_dir().join(self._actual_path)
        self._ignore_lines = kwargs.get("ignore_lines")
        self.j2_ctx = kwargs.get("j2_ctx", PKDict())
        self.is_bytes = kwargs.get("is_bytes", False)


def _base_dir(postfix):
    """Base name with directory.

    Args:
        postfix (str): what to append to base (``_data`` or ``_work``).

    Returns:
        py.path.local: base directory with postfix
    """
    f = _test_file()
    if not f:
        raise PKFail("unable to find test file path; not running in pykern.pkcli.test?")
    b = re.sub(r"_test$|^test_", "", f.purebasename)
    assert b != f.purebasename, "{}: module name must end in _test".format(f)
    return f.new(basename=b + postfix).realpath()


def _fail(std_message_args, *args, **kwargs):
    """Augment standard failure messages with args and kwargs

    Args:
        std_message_args (tuple): fmt string and args. eg. ("expect={} != actual={}", expect, actual)
    """
    if args:
        pkfail(
            f"{std_message_args[0]} {args[0]}",
            *std_message_args[1:],
            *args[1:],
            **kwargs,
        )
    pkfail(*std_message_args)


def _pkdlog(*args, **kwargs):
    from pykern.pkdebug import pkdlog

    pkdlog(*args, **kwargs)


def _test_file():
    """Various ways to initialize _test_file"""
    global _init_test_file, __test_file

    if not _init_test_file:
        _init_test_file = True
        # pykern.pkcli.test
        t = os.environ.get(TEST_FILE_ENV)
        if t:
            __test_file = py.path.local(t)
    if __test_file:
        return __test_file
    if module_under_test:
        # POSIT: pykern.pytest_plugin or sirepo/tests/conftest.py
        m = module_under_test
        return py.path.local(m.__file__)
    # py.test alone, just guess
    s = inspect.currentframe().f_back.f_back
    f = None
    for _ in range(100):
        if s.f_code.co_filename.endswith("_test.py"):
            return py.path.local(s.f_code.co_filename)
        s = s.f_back
        if not s:
            break
    return None



"""run test files in separate processes

:copyright: Copyright (c) 2019 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

from pykern import pkconfig
from pykern import pkio
from pykern import pksubprocess
from pykern import pkunit
from pykern.pkcollections import PKDict
from pykern.pkdebug import pkdp
import os
import pykern.pkcli
import re
import sys


SUITE_D = "tests"
_COROUTINE_NEVER_AWAITED = re.compile(
    "(.+ coroutine \S+ was never awaited.*)", flags=re.MULTILINE
)
_TEST_SKIPPED = re.compile(r"^.+\s+SKIPPED\s+\(.+\)$", flags=re.MULTILINE)
_TEST_PY = re.compile(r"_test\.py$")


_cfg = pkconfig.init(
    ignore_warnings=(False, bool, "override pytest's output of all warnings"),
    max_failures=(5, int, "maximum number of test failures before exit"),
    restartable=(
        False,
        bool,
        "allow pkunit.restart_or_fail to restart",
    ),
)


def default_command(*args):
    """Run tests one at a time with py.test.

    Searches in ``tests`` sub-directory if not provided a
    list of tests or not in tests/ already

    Arguments are directories or files, which are searched for _test.py
    files.

    An argument which is ``case=<pattern>``, is passed to pytest
    as ``-k <pattern>``.

    ``skip_past=<last_to_ignore>`` causes collection to ignore all
    files up to and including ``<last_to_ignore>``` (may be partial
    match of the test file).

    Writes failures to ``<base>_test.log``

    Args:
        args (str): test dirs, files, options
    Returns:
        str: passed=N if all passed, else raises `pkcli.Error`

    """
    return _Test(args).result


class _Test:
    def __init__(self, args):
        self.count = 0
        self.failures = []
        self._args(args)
        for t in self.paths:
            self.count += 1
            self._run_one(t)
            if len(self.failures) >= _cfg.max_failures:
                sys.stdout.write(
                    f"too many failures={len(self.failures)} aborting\n",
                )
                break
        if self.count == 0:
            pykern.pkcli.command_error("no tests found")
        self._assert_failures()
        self.result = f"passed={self.count}"

    def _args(self, tests):
        def _file(path):
            if self.skip_past:
                if self.skip_past in path:
                    self.skip_past = None
                return
            self.paths.append(path)

        def _find(paths):
            i = re.compile(r"(?:_work|_data)/")
            cwd = pkio.py_path()
            for t in _resolve_test_paths(paths, cwd):
                t = pkio.py_path(t)
                if t.check(file=True):
                    _file(str(cwd.bestrelpath(t)))
                    continue
                for p in pkio.walk_tree(t, _TEST_PY):
                    p = str(cwd.bestrelpath(p))
                    if not i.search(p):
                        _file(p)

        def _flag(name, value):
            if len(value) <= 0:
                pykern.pkcli.command_error(f"empty value for option={name}")
            elif name == "case":
                self.flags.extend(("-k", value))
            elif name == "skip_past":
                if self.skip_past:
                    pykern.pkcli.command_error(
                        f"skip_past={value} passed twice (skip_past={self.skip_past})"
                    )
                self.skip_past = value
            else:
                pykern.pkcli.command_error(f"unsupported option={name}")

        def _resolve_test_paths(paths, current_dir):
            if not paths:
                p = current_dir
                if p.basename != SUITE_D:
                    p = SUITE_D
                paths = (p,)
            return paths

        p = []
        self.flags = []
        self.skip_past = None
        for t in tests:
            if "=" in t:
                _flag(*(t.split("=")))
            else:
                p.append(t)
        self.paths = []
        _find(p)

    def _assert_failures(self):
        if len(self.failures) <= 0:
            return
        # Avoid dumping too many test logs
        for o in self.failures[:5]:
            sys.stdout.write(pkio.read_text(o))
        sys.stdout.flush()
        pykern.pkcli.command_error(
            f"FAILED={len(self.failures)} passed={self.count - len(self.failures)}",
        )

    def _remove_work_dir(self, test_file):
        w = _TEST_PY.sub(pkunit.WORK_DIR_SUFFIX, test_file)
        if w != test_file:
            pkio.unchecked_remove(w)

    def _run_one(self, test_f):
        def _env(restartable):
            res = os.environ.copy()
            res.update(
                {
                    pkunit.TEST_FILE_ENV: str(pkio.py_path(test_f)),
                    pkunit.RESTARTABLE: "1" if restartable else "",
                }
            )
            return res

        def _except(exc, output, restartable):
            if isinstance(exc, RuntimeError):
                if "exit(5)" in exc.args[0]:
                    # 5 means test was skipped
                    # see http://doc.pytest.org/en/latest/usage.html#possible-exit-codes
                    return "skipped"
                if t and "exit(2)" in exc.args[0]:
                    # 2 means KeyboardInterrupt
                    # POSIT: pkunit.restart_or_fail uses this
                    return "restart"
            return _fail(output)

        def _fail(output):
            self.failures.append(output)
            return f"FAIL {output}"

        def _ignore_warnings():
            if not _cfg.ignore_warnings:
                return []
            rv = []
            for w in (
                # https://docs.python.org/3/library/warnings.html#default-warning-filter
                "DeprecationWarning",
                "PendingDeprecationWarning",
                "ImportWarning",
                "ResourceWarning",
            ):
                rv.extend(("-W", f"ignore::{w}"))
            return rv

        def _try(output, restartable):
            sys.stdout.write(test_f)
            sys.stdout.flush()
            c = (
                ["pytest"]
                + _ignore_warnings()
                + [
                    "--tb=native",
                    "-v",
                    "-s",
                    "-rs",
                    test_f,
                ]
                + self.flags
            )
            v = _env(restartable)
            try:
                pksubprocess.check_call_with_signals(c, output=output, env=v)
            except Exception as e:
                return _except(e, output, restartable)
            o = pkio.read_text(output)
            if m := re.findall(_COROUTINE_NEVER_AWAITED, o):
                with pkio.py_path(output).open(mode="a") as f:
                    f.write("".join(f"ERROR: {x}\n" for x in m))
                return _fail(output)
            if _TEST_SKIPPED.search(o):
                return "\n".join(["pass"] + _skipped(o))
            return "pass"

        def _skipped(ouput):
            return _TEST_SKIPPED.findall(ouput)

        o = test_f.replace(".py", ".log")
        for t in range(4 if _cfg.restartable else 0, -1, -1):
            self._remove_work_dir(test_f)
            pkio.unchecked_remove(o)
            m = _try(o, t > 0)
            sys.stdout.write(" " + m + "\n")
            if m != "restart":
                return



"""run github backups and restores

:copyright: Copyright (c) 2013-2018 Bivio Software, Inc.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

from pykern import pkcli
from pykern import pkconfig
from pykern import pkio
from pykern import pkjson
from pykern.pkcollections import PKDict
from pykern.pkdebug import pkdlog, pkdp, pkdc, pkdexc
import datetime
import github3
import github3.exceptions
import glob
import json
import pykern.pkcollections
import os
import os.path
import re
import subprocess
import sys
import time


_GITHUB_HOST = "github.com"
_GITHUB_URI = "https://" + _GITHUB_HOST
_GITHUB_API = "https://api." + _GITHUB_HOST
_WIKI_ERROR_OK = r"fatal: repository 'https://github.com/[-/\w\.]+.wiki.git/' not found"
_RE_TYPE = type(re.compile(""))
_MAX_TRIES = 3
_TEST_REPOS = [
    PKDict(org="radiasoft", name="test-pykern-github"),
    PKDict(org="radiasoft", name="test-pykern-github-no-wiki"),
    PKDict(org="biviosoftware", name="test-pykern-github-no-wiki"),
]
_TXZ = ".txz"
_LIST_ARG_SEP_RE = re.compile(r"[\s,:;]+")


class GitHub(object):
    def __init__(self):
        self._github = None

    @classmethod
    def indent2(cls, text):
        return re.sub(r"^(?=[^\n])", "  ", text, flags=re.MULTILINE)

    @classmethod
    def issue_body(cls, issue):
        res = issue.body
        if res is None or len(res) == 0:
            return ""
        res = res.replace("\r", "")
        if not res.endswith("\n"):
            res += "\n"
        return res

    def login(self):
        self._github = (
            github3.GitHub(username=cfg.user, password=cfg.password)
            if cfg.password
            else github3.GitHub()
        )
        return self._github

    def milestone(self, repo, title):
        t = title.lower()
        r = self.repo_arg(repo)
        for m in r.milestones(state="open"):
            if m.title.lower() == t:
                return m.number
        raise KeyError(f"milestone={title} not found in repo={r.name}")

    def repo(self, repo):
        if not self._github:
            self.login()
        a = repo.split("/")
        if len(a) == 1:
            a.insert(0, "radiasoft")
        return self._github.repository(*a)

    def repo_arg(self, repo):
        if not repo:
            pkcli.command_error("repo argument not supplied")
        return self.repo(repo) if isinstance(repo, str) else repo

    def list_org_repos(self, org, include_forks):
        """Returns list of repos for org"""
        self.login()
        res = []
        for r in self._github.organization(org).repositories():
            if include_forks or not r.fork:
                res.append(r)
        return res


def backup(org):
    """Backs up all github repos in org into pwd

    Args:
        org (str): backup all repos visible to user

    Creates timestamped directory, and purges directories older than cfg.keep_days

    Note: backups are incremental in case of repositories to save
    space. Hard-linking is used between "keep_dates". The assumption
    here is that the github backup is actually copied to a backup
    server.
    """
    try:
        _Backup(org)
    except subprocess.CalledProcessError as e:
        if hasattr(e, "output"):
            pkdlog("ERROR: Backup {}", e.output)
    pkdlog("DONE")


def ci_check(repo, branch=None):
    def _branch(r, name, reraise=True):
        try:
            return r.branch(name=name)
        except github3.exceptions.NotFoundError:
            if reraise:
                raise
            return None

    r = GitHub().repo_arg(repo)
    b = (
        _branch(r, branch)
        if branch
        else _branch(r, "master", False) or _branch(r, "main")
    )
    s = b.commit.sha
    c = [c.conclusion for c in b.commit.check_runs()]
    i = f"repo={repo} branch={b.name} sha={s}"
    if not c:
        pkcli.command_error(f"{i} No workflow runs for commit")
    if c[0] != "success":
        pkcli.command_error(f"{i} Unsuccessful conclusion={c}")
    return f"{i} Passed CI"


def collaborators(org, filename, affiliation="outside", private=True):
    """Lists direct repos to which user has access and is not a team member

    Configured user must be an owner

    Args:
        org (str): GitHub organization
        affiliation (str): all, direct, outside
    """
    from pykern import pkyaml

    g = GitHub().login()
    o = g.organization(org)
    res = dict()
    for r in o.repositories():
        if r.private == private:
            x = [str(c.login) for c in r.collaborators(affiliation=affiliation)]
            if x:
                res[str(r.name)] = x
    pkyaml.dump_pretty(res, filename)


def create_issue(repo, title, body="", assignees=None, labels=None, milestone=None):
    g = GitHub()
    r = g.repo_arg(repo)
    a = PKDict()
    if milestone:
        try:
            m = int(milestone)
            assert m > 0
            a.milestone = str(m)
        except Exception:
            a.milestone = g.milestone(r, title=milestone)

    def _list_arg(arg):
        if isinstance(arg, str):
            return _LIST_ARG_SEP_RE.split(arg)
        return arg

    if labels:
        a.labels = _list_arg(labels)
    if assignees:
        a.assignees = _list_arg(assignees)
    return r.create_issue(
        title=title,
        body=body,
        **a,
    ).number


def create_milestone(repo, title, description="", due_on=None):
    a = PKDict()
    if due_on:
        # GitHub seems to always create at 08:00:00Z in any zone
        # so just have the user put in the time.
        a.due_on = due_on + "T08:00:00Z"
    if description:
        a.description = description
    return (
        GitHub()
        .repo_arg(repo)
        .create_milestone(
            title=title,
            **a,
        )
        .number
    )


def get_milestone(repo, title):
    return GitHub().milestone(repo, title)


def issue_pending_alpha(repo):
    """Create "Alpha Release [pending]" issue

    This should be created after the current alpha completes.
    """
    r, a = _alpha_pending(repo, assert_exists=False)
    if a:
        return f"#{a.number} {a.title} already exists"
    i = _create_release_issue(r, _release_title("Alpha", pending=True), "")
    return f"Created #{i.number}"


def issue_start_alpha(repo):
    r, a = _alpha_pending(repo)
    if not a.body:
        raise ValueError("no new issues before prior release so no need to start alpha")
    for i in r.issues(state="open"):
        if i.number != a.number and re.search(
            r"^alpha release \d+", i.title, flags=re.IGNORECASE
        ):
            _assert_closed(i)
            # does not get here
    a.edit(title=_release_title("Alpha"))
    return f"Started #{a.number} and {issue_pending_alpha(repo)}" + (
        a.body if cfg.test_mode else ""
    )


def issue_start_beta(repo):
    return _promote(repo, "Alpha", "Beta")


def issue_start_prod(repo):
    return _promote(repo, "Beta", "Prod")


def issue_update_alpha_pending(repo):
    r, a = _alpha_pending(repo)
    res = ""
    b = a.body or ""
    # the loop below picks up the pending so don't add b to p
    p = []
    for i in r.issues(state="all", sort="updated", direction="desc"):
        if re.search(f"^alpha release", i.title, flags=re.IGNORECASE):
            p.append(i.body or "")
            # somewhat arbitrary
            if len(p) > 10:
                break
    p = "\n".join(p)
    g = GitHub()
    g.login()
    for c in r.commits(
        sha="master",
        since=datetime.datetime.now() - datetime.timedelta(minutes=24 * 60),
    ):
        m = re.search(r"([-\w]+/[-\w]+)?#(\d+)", c.message)
        if not m:
            res += f"commit={c.sha} missing #NN in message={c.message}, ignoring\n"
            continue
        n = m.group(1) or r.full_name
        try:
            i = g.repo(n).issue(m.group(2))
        except Exception as e:
            res += f"Issue {n}#{m.group(2)} exception={e}\n"
            continue
        z = f"\\b{n}#{i.number}"
        if n == r.full_name:
            z += "|#{i.number}"
        y = re.compile("(?:" + z + r")\b")
        if y.search(p):
            # don't bother to note already included commits; also makes
            # unit test simpler
            continue
        if b and not b.endswith("\n"):
            b += "\n"
        x = f"- {i.title} {n}#{i.number}\n"
        b += x
        a.edit(body=b)
        res += f"Updated #{a.number} with: {x}"
    return res


def issues_as_csv(repo):
    """Export issues as CSV

    Args:
        repo (str): will add radiasoft/ if missing
    """
    cols = (
        "number",
        "title",
        "assignees",
        "comments_count",
        "comments_url",
        "created_at",
        "events_url",
        "html_url",
        "id",
        "labels_url",
        "locked",
        "milestone",
        "original_labels",
        "pull_request_urls",
        "state",
        "updated_at",
        "user",
        "body",
    )

    def _s(v):
        if v is None:
            return ""
        if isinstance(v, list):
            return ",".join([_s(x) for x in v])
        return str(getattr(v, "name", v))

    specials = set("\n,")

    def _c(i, c):
        v = _s(getattr(i, c)).replace('"', '""')
        if any(c in specials for c in v):
            return f'"{v}"'
        return v

    r = GitHub().repo_arg(repo)
    n = r.name + ".csv"
    with open(n, mode="w") as f:

        def _write(v):
            # Need custom csv, because utf8 not handled by py2's csv
            f.write(",".join(v) + "\r\n")

        _write(cols)
        for i in r.issues(state="open"):
            _write([_c(i, c) for c in cols])
    return n


def labels(repo, clear=False):
    """Setup the RadiaSoft labels for ``repo``.

    Will add "radiasoft/" to the name if it is missing.

    Args:
        repo (str): will add https://github.com/radiasoft if missing
        clear (bool): if True, clear all existing labels
    """
    r = GitHub().repo_arg(repo)
    if clear:
        for l in r.labels():
            l.delete()
    for x in (
        ("customer", "0e8a16"),
        ("devops", "84b6eb"),
        ("doc", "84b6eb"),
        ("question", "84b6eb"),
        ("release", "84b6eb"),
        ("sw", "84b6eb"),
        ("test", "84b6eb"),
        ("user", "0e8a16"),
    ):
        try:
            r.create_label(*x)
        except github3.exceptions.UnprocessableEntity:
            # 422 Validation Failed: happens because already exists
            pass


def list_repos(org, include_forks=False):
    """Lists repos for org, possibly including forks

    Args:
        org (str): GitHub organization
        include_forks (bool): include forks or not
    """
    return sorted(
        (str(r.name) for r in GitHub().list_org_repos(org, include_forks=include_forks))
    )


def restore(git_txz):
    """Restores the git directory (only) to a new directory with the .git.txz suffix"""
    m = re.search(r"(([^/]+)\.git)\.txz$", git_txz)
    if not m:
        raise ValueError(git_txz, ": does not end in .git.txz")
    git_txz = pkio.py_path(git_txz)
    d = m.group(2)
    pkdc("restore: {}", d)
    g = m.group(1)
    with pkio.save_chdir(d, mkdir=True):
        _shell(["tar", "xJf", str(git_txz)])
        os.rename(g, ".git")
        _shell(["git", "config", "core.bare", "false"])
        _shell(["git", "config", "core.logallrefupdates", "true"])
        _shell(["git", "checkout"])


class _Backup(GitHub):
    def __init__(self, org):
        def _repos():
            if cfg.test_mode:
                self.login()
                return [self._github.repository(r.org, r.name) for r in _TEST_REPOS]
            return _try(lambda: self.list_org_repos(org, include_forks=True))

        # POSIT: timestamps are sorted in _clone()
        self._date_d = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
        with pkio.save_chdir(self._date_d, mkdir=True):
            for r in _repos():
                pkdlog("{}: begin", r.full_name)
                self._repo(r)
        self._purge()

    def _extract_backup(self, backup):
        pkdc("updating from {}", backup)
        _shell(["tar", "xJf", str(backup)])

    def _prev_backup(self, base, ext):
        # POSIT: timestamp Backup
        b = pkio.sorted_glob(f"../*/{base}{ext}")
        return b[-1] if b else []

    def _purge(self):
        expires = datetime.datetime.utcnow() - cfg.keep_days
        for d in pkio.sorted_glob("[0-9]" * len(self._date_d)):
            t = datetime.datetime.utcfromtimestamp(d.stat().mtime)
            if t < expires:
                pkio.unchecked_remove(d)

    def _repo(self, repo):
        fn = repo.full_name
        bd = re.sub("/", "-", fn)

        def _clone(suffix):
            base = bd + suffix
            prev = self._prev_backup(base, ext="")
            if not prev:
                _shell(
                    (
                        "git",
                        "clone",
                        "--quiet",
                        "--mirror",
                        _GITHUB_URI + "/" + fn + suffix,
                        base,
                    ),
                )
                return
            _shell(("cp", "--archive", "--link", str(prev), "./"))
            with pkio.save_chdir(base):
                l = pkio.py_path("gc.log")
                if l.check():
                    pkdlog("gc.log={}", pkio.read_text(l))
                    pkio.unchecked_remove(l)
                _shell(["git", "remote", "update", "--prune"])

        def _issues():
            def _issue(i, d):
                j = _trim_body(i)
                j["comments"] = [_trim_body(c) for c in i.comments()]
                p = i.pull_request()
                if p:
                    j["review_comments"] = [_trim_body(c) for c in p.review_comments()]
                pkjson.dump_pretty(j, filename=d.join(str(i.number) + ".json"))

            if not repo.has_issues:
                return
            base = bd + ".issues"
            prev = self._prev_backup(base, ext=_TXZ)
            d = pkio.mkdir_parent(base)
            k = PKDict(state="all")
            if prev:
                self._extract_backup(prev)
                k.since = datetime.datetime.now() - datetime.timedelta(days=7)
            for i in _try(lambda: list(repo.issues(**k))):
                _try(lambda: _issue(i, d))
            _tar(base)

        def _json(gen, suffix):
            base = bd + suffix
            with open(base, "wt") as f:
                sep = "[\n"
                for i in gen:
                    f.write(sep)
                    f.write(pkjson.dump_pretty(_trim_body(i)))
                    sep = ",\n"
                if "[" in sep:
                    # Empty iteration
                    f.write(sep)
                f.write("]\n")
            _shell(["xz", base])

        def _tar(base):
            _shell(["tar", "cJf", base + _TXZ, base])
            pkio.unchecked_remove(base)

        def _trim_body(o):
            res = o.as_dict()
            try:
                # github returns three formats, and we only want source
                del res["body_text"]
                del res["body_html"]
            except KeyError:
                pass
            return res

        try:
            # backup the code first; should be fast
            _clone(".git")
            _issues()
            if repo.has_wiki:
                try:
                    _clone(".wiki.git")
                except subprocess.CalledProcessError as e:
                    if not re.search(_WIKI_ERROR_OK, str(e.output)):
                        raise
            _try(lambda: _json(repo.comments(), ".comments"))
            # TODO(robnagler) releases, packages, projects
            return
        except Exception as e:
            pkdlog(
                "ERROR: {} {} {} {} {}",
                fn,
                type(e),
                e,
                getattr(e, "output", None),
                pkdexc(),
            )


def _alpha_pending(repo, assert_exists=True):
    r = GitHub().repo(repo)
    for a in list(r.issues(state="open")):
        if re.search(r"^alpha release.*pending", a.title, flags=re.IGNORECASE):
            return r, a
    assert not assert_exists, '"Alpha Release [pending]" issue not found'
    return r, None


def _assert_closed(issue):
    assert issue.state == "closed", f"Need to close #{issue.number} {issue.title}"


def _cfg():
    global cfg
    n = None
    p = PKDict(
        api_pause_seconds=(
            0 if pkconfig.in_dev_mode() else 10,
            int,
            "pauses between backups",
        ),
        exclude_re=(None, _cfg_exclude_re, "regular expression to exclude a repo"),
        keep_days=(
            _cfg_keep_days(2),
            _cfg_keep_days,
            "how many days of backups to keep",
        ),
        password=[None, str, "github passsword"],
        test_mode=(
            pkconfig.in_dev_mode(),
            pkconfig.parse_bool,
            f"only backs up {_TEST_REPOS} repos",
        ),
        user=[None, str, "github user"],
    )
    cfg = pkconfig.init(**p)
    assert (
        cfg.test_mode or cfg.password is not None and cfg.user is not None
    ), "user and password required unless test_mode"


def _cfg_exclude_re(anything):
    if isinstance(anything, _RE_TYPE):
        return anything
    return re.compile(anything, flags=re.IGNORECASE)


def _cfg_keep_days(anything):
    if isinstance(anything, datetime.timedelta):
        return anything
    return datetime.timedelta(days=int(anything))


def _create_release_issue(repo, title, body):
    return repo.create_issue(title=title, body=body, labels=["release"])


def _promote(repo, prev, this):
    r = GitHub().repo_arg(repo)
    b = ""
    for i in r.issues(state="all", sort="updated", direction="desc"):
        if re.search(f"^{this} release", i.title, flags=re.IGNORECASE):
            _assert_closed(i)
            t = i
            break
        if re.search(f"^{prev} release", i.title, flags=re.IGNORECASE):
            if "pending" in i.title:
                continue
            _assert_closed(i)
            b += f"- #{i.number} {i.title}\n" + GitHub.indent2(GitHub.issue_body(i))
    else:
        raise AssertionError(f'No previous "{this} Release" issue found')
    assert b, f'no "{prev} Release" found, since #{t.number} {t.title}'
    i = _create_release_issue(r, _release_title(this), b)
    return f"Created #{i.number} {i.title}" + (b if cfg.test_mode else "")


def _release_title(channel, pending=False):
    x = (
        "[pending]"
        if pending
        else datetime.datetime.utcnow()
        .replace(
            microsecond=0,
        )
        .isoformat(sep=" ")
        + " UTC"
    )
    return f"{channel} Release {x}"


def _shell(cmd):
    subprocess.check_output(cmd, stderr=subprocess.STDOUT)


def _try(op):
    for t in range(_MAX_TRIES, 0, -1):
        try:
            return op()
        except github3.exceptions.ForbiddenError as e:
            x = getattr(e, "response", None)
            # Response ojects return false so don't use "not x" here
            if x is None:
                pkdlog(
                    'no "response" in ForbiddenError attributes={}',
                    [(a, getattr(e, a)) for a in dir(e) if not a.startswith("_")],
                )
                raise
            h = getattr(x, "headers", None)
            # See above. Being cautious about falsey testing
            if h is None:
                pkdlog('no "headers" in ForbiddenError response={}', h)
                raise
            r = h.get("X-RateLimit-Remaining", "n/a")
            if r != "0":
                pkdlog("some other error(?) X-RateLimit-Remaining={}", r)
                raise
            if t == 0:
                pkdlog("MAX_TRIES={} reached", _MAX_TRIES)
                raise
            r = int(h["X-RateLimit-Reset"])
            n = int(time.time())
            s = r - n
            if s <= 0:
                pkdlog("trying min sleep X-RateLimit-Reset={} <= now={}", r, n)
                s = 60
            elif s > 4000:
                # Should reset in an hour if the GitHub API is right
                pkdlog("trying max sleep; X-RateLimit-Reset={} > 4000 + now={}", r, n)
                s = 3600
            pkdlog("RateLimit hit sleep={}", s)
            time.sleep(s)
    raise AssertionError("should not get here")


_cfg()



# -*- coding: utf-8 -*-
"""test pkjson

:copyright: Copyright (c) 2017 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""


def test_load_any():
    import json
    from pykern import pkcollections
    from pykern import pkjson
    from pykern.pkunit import pkeq

    j = json.dumps(["a", "b"])
    j2 = pkjson.load_any(j)
    pkeq("a", j2[0])
    j3 = json.dumps({"a": 33})
    j4 = pkjson.load_any(j3)
    pkeq(
        33,
        j4.a,
        "{}: j4.a is not 33",
        j4.a,
    )
    j5 = json.dumps({"a": 33, "b": {"values": "will collide, but ok"}})
    j6 = pkjson.load_any(j5)
    pkjson.load_any(j5, object_pairs_hook=pkcollections.PKDict)


def test_load_any_parse_int():
    from pykern import pkjson
    from pykern.pkunit import pkeq, pkexcept

    for c, e in (
        ("9007199254740992", float(2**53)),
        ("9007199254740991", int(2**53 - 1)),
        ("-9007199254740991", int(-(2**53) + 1)),
        ("-9007199254740992", float(-(2**53))),
        ("1" * 64, 1.1111111111111112e63),
    ):
        a = pkjson.load_any(c)
        pkeq(type(e), type(a))
        pkeq(e, a)
    with pkexcept("unreasonably large"):
        pkjson.load_any("1" * 65)


def test_dump_bytes():
    import json
    from pykern import pkjson, pkcompat
    from pykern.pkunit import pkeq

    v = ["a", "b"]
    expect = pkcompat.to_bytes(
        json.dumps(v).replace(" ", ""),
    )
    actual = pkjson.dump_bytes(v)
    pkeq(expect, actual)
    actual = pkjson.load_any(actual)
    pkeq(v, actual)


def test_dump_pretty():
    from pykern import pkjson
    from pykern.pkunit import pkeq

    class Other(object):
        def __init__(self, x):
            self.x = x

        def __str__(self):
            return str(self.x)

    v = {"d": ["a", "b"], "c": Other("xyz")}
    a = pkjson.dump_pretty(v)
    pkeq(
        """{
    "c": "xyz",
    "d": [
        "a",
        "b"
    ]
}
""",
        a,
    )



# -*- coding: utf-8 -*-
"""?

:copyright: Copyright (c) 2015 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from __future__ import absolute_import, division, print_function
import pytest


def test_run_path_as_module():
    import sys
    from pykern import pkunit
    from pykern import pkrunpy

    m = pkrunpy.run_path_as_module(pkunit.data_dir().join("f1.py"))
    assert (
        m.func1() == sys.modules
    ), "When imported, should be able to call function within module"



from __future__ import absolute_import, division, print_function

last_cmd = None

from pykern.pkdebug import pkdp

def cmd1(arg1):
    """Subject line for cmd1

    Args:
        arg1
    """
    global last_cmd
    last_cmd = cmd1
    return

def cmd2():
    """Subject line for cmd2

    Args:
        -
    """
    global last_cmd
    last_cmd = cmd2
    return
